/*********************************************************************
; Copyright 1992-97, The University of Iowa (UI).  All rights reserved. 
; By using this software the USER indicates that he or she has read, 
; understood and will comply with the following:
;
; --- UI hereby grants USER nonexclusive permission to use, copy and/or
; modify this software for internal, noncommercial, research purposes only.
; Any distribution, including commercial sale or license, of this software,
; copies of the software, its associated documentation and/or modifications
; of either is strictly prohibited without the prior consent of UI.  Title
; to copyright to this software and its associated documentation shall at
; all times remain with UI.  Appropriate copyright notice shall be placed
; on all software copies, and a complete copy of this notice shall be
; included in all copies of the associated documentation.  No right is
; granted to use in advertising, publicity or otherwise any trademark,
; service mark, or the name of UI.  Software and/or its associated
; documentation identified as "confidential," if any, will be protected
; from unauthorized use/disclosure with the same degree of care USER
; regularly employs to safeguard its own such information.
;
; --- This software and any associated documentation is provided "as is," and
; UI MAKES NO REPRESENTATIONS OR WARRANTIES, EXPRESS OR IMPLIED, INCLUDING
; THOSE OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE, OR THAT
; USE OF THE SOFTWARE, MODIFICATIONS, OR ASSOCIATED DOCUMENTATION WILL
; NOT INFRINGE ANY PATENTS, COPYRIGHTS, TRADEMARKS OR OTHER INTELLECTUAL
; PROPERTY RIGHTS OF A THIRD PARTY.  UI, the University of Iowa,
; its Regents, officers, and employees shall not be liable under any
; circumstances for any direct, indirect, special, incidental, or
; consequential damages with respect to any claim by USER or any third
; party on account of or arising from the use, or inability to use, this
; software or its associated documentation, even if UI has been advised
; of the possibility of those damages.
*********************************************************************/
#include "main.h"
#include "trie.h"

#define EXTRA_SPACE (GROW<<2)
#define HAVE_UIP 0
#define PURECHECK -MAX_ATOM
#define MAGIC 7 /* at most 12*/
#define two14 16383   /* 2^14-1 */
#define two15 32767   /* 2^15-1 */

int Tmp_idx_end = MAX_ATOM;
struct trie *Conflict_cl, *Old_cls, *BItrie;
struct trie *Nonunit_cls = NULL;
int Conflict_cl_level;
long Pure_num;
int UIP, UIP_num, Jump_num, NH_num, Miss_num;
int Max_cand;

#define trie_assign_value(i, val)\
  if (Backup_idx) { V3[i]->next = Top_var3->next; Top_var3->next = V3[i]; }\
  assign_value(i, val)

#define trie2_assign_value(i, val)\
  if (Backup_idx) { V2[i]->next = Top_var2->next; Top_var2->next = V2[i]; }\
  assign_value(i, val)

int bugnum = 0;

void trie_init_once_forall ()
{
  if (sizeof(struct trie) % sizeof(int *)) 
    printf("\nWARNING: sizeof(struct trie) = %d\n", sizeof(struct trie));
  if (sizeof(struct var2) % sizeof(int *)) 
    printf("\nWARNING: sizeof(struct var2) = %d\n", sizeof(struct var2));
  if (sizeof(struct var3) % sizeof(int *)) 
    printf("\nWARNING: sizeof(struct var3) = %d\n", sizeof(struct var3));
  if (sizeof(struct cap) % sizeof(int *)) 
    printf("\nWARNING: sizeof(struct cap) = %d\n", sizeof(struct cap));

  V2 = NULL;
  V3 = NULL;
  CLtrie = NHtrie = Root = NULL;
  Trie_news = Trie_frees = Trie_gets = Trie_avails = 0;
  Trie_avail = NULL;
  BItrie = CLAUSE3 = get_trie();
}

void init_trie ()
{
  free_trie(Root); 
  free_trie(NHtrie); 
  free_trie(CLtrie);  
  Old_cls = NULL;
  Root = CLtrie = NHtrie = NULL;
  Backup_idx = 0;
  Top_var2 = NULL;
  Top_var3 = NULL;
  NH_num = Jump_num = Pure_num = UIP_num = Clause_extra_num = 0;
  Miss_num = 0;
}

struct trie *trie_create_one_path (arr, signs, count)
     int arr[], signs[], count;
{
  int sign;
  struct trie *first, *pre, *no;

  first = no = get_trie(); 
  no->key = arr[count]; 
  sign = signs[no->key];
  while (--count >= 0) {
    pre = no;
    no = get_trie();
    if (sign) pre->pos = no;
    else pre->neg = no;
    no->key = arr[count]; 
    sign = signs[no->key];
  }
  if (sign)
    no->pos = CLAUSE3;
  else
    no->neg = CLAUSE3;
  return first;
}

int trie_insert_clause ( cl_arr, sign_arr, total )             
     int cl_arr[], sign_arr[], total;
{
  if (Root == CLAUSE3) return 1;

  if (total > 0) 
    Root = trie_insert_clause2(Root, cl_arr, sign_arr, total-1);

  if (total == 0 || Root == CLAUSE3) { /* an empty clause is found */

#ifdef MORETRACE
    if (TRACE > 6)
      fprintf(Sato_fp, "    [C%ld] becomes empty.\n", Clause_num);
#endif

    bug(4);
    free_trie(Root);
    Root = CLAUSE3;
    return 1;
  }
  /*print_trie(Root);*/
  return 0;
}

struct trie *trie_insert_clause2 (root, arr, signs, count)
     struct trie *root; /* current root of the tree */
     int arr[], signs[]; /* current clause */
     int count; /* the length of the current clause */
{
  int key;
  int sign, dep;
  struct trie *no, *bro1, *bro2, *pars[MAX_LENGTH];
  static struct trie *roots[MAX_ATOM];

  if (root == NULL)
    for (dep = 0; dep <= Max_atom; dep++) roots[dep] = NULL;

  key = arr[count];

  if (roots[key] == NULL) {

    if (GROW && Backup_idx) {
      for (dep = key-1; dep >= 0; dep--) {
	if (roots[dep] != NULL) {
	  return trie2_insert_clause(roots[dep], arr, signs, count);
	}
      }
      return trie2_insert_clause(root, arr, signs, count);
    }


    roots[key] = no = trie_create_one_path (arr, signs, count);
    for (dep = key-1; dep >= 0; dep--) {
      if (roots[dep] != NULL) {
	no->bro = roots[dep]->bro;
	roots[dep]->bro = no;
	return root;
      }
    }
    no->bro = root;
    return no;
  }

  no = roots[key];

  if (GROW && Backup_idx) 
    return trie2_insert_clause(no, arr, signs, count);

  dep = 0;
  bro1 = bro2 = NULL;

  while (no != NULL && count >= 0) {
    key = arr[count];
    sign = signs[key]; 
    
    if (key == no->key) {
      if (sign == 1) {

	if HAS_POS_CL(no) {
#ifdef MORETRACE
	  print_x_subsume;
#endif
	  Subsume_num++;
	  return root; /* skip this clause */

	} else if (trie_subsume_clause(no->neg, arr, signs, count)) {
#ifdef MORETRACE
	  print_x_skip;
#endif
	  count--;
	  bro1 = no;
	  no = no->bro; 
	} else {
	  count--;
	  pars[dep++] = no;
	  bro1 = NULL;
	  if (sign) no = no->pos; else no = no->neg;
	}
      } else { /* sign == 0 */

	if HAS_NEG_CL(no) {
#ifdef MORETRACE
	  print_x_subsume;
#endif
	  Subsume_num++;
	  return root; /* skip this clause */

	} else if (trie_subsume_clause(no->pos, arr, signs, count)) {
#ifdef MORETRACE
	  print_x_skip;
#endif
	  count--;
	  bro1 = no;
	  no = no->bro; 
	} else {
	  count--;
	  pars[dep++] = no;
	  bro1 = NULL;
	  if (sign) no = no->pos; else no = no->neg;
	}
      }
    } else if (key < no->key) {
      bro2 = no;
      no = NULL;
    } else { /* (key > no->key) */
      bro1 = no;
      no = bro1->bro;
      while ((no != NULL) && (no->key < key)) {
	bro1 = no;
	no = bro1->bro;
      }
    }
  }

  if (count >= 0) {

    if (count == 0 && dep == 0) { /* a unit clause is found */ 
      key = arr[0];
      assign_value(key, signs[key]);
      Unit_num++;
      if (OUTPUT) print_unit_clause(Sato_fp, Value[key], key);

#ifdef MORETRACE
      if (TRACE > 7 || TRACE == 4)
	fprintf (Sato_fp, "    x%d is set to %d by unit clause.\n",
		 key, Value[key]);
#endif
      return root;
    }

    no = trie_create_one_path (arr, signs, count);
    no->bro = bro2;
    if (bro1 == NULL) {
      if (dep--) {
	if (signs[pars[dep]->key])
	  pars[dep]->pos = no;
	else 
	  pars[dep]->neg = no;
	return root;
      } else {
	return no;
      }
    } else {
      bro1->bro = no;
      if (dep == 0) roots[no->key] = no;
      return root;
    }
  }

  /* count < 0 */
  while (dep--) {
    no = pars[dep];
    sign = signs[no->key];

    if (dep == 0) { /* if at depth 0, no is unit clause */
      key = no->key;
      assign_value(key, sign);
      Unit_num++;

#ifdef MORETRACE
      if (TRACE > 7 || TRACE == 4)
	fprintf (Sato_fp, "    %sx%d is found be a unit clause.\n",
		 (sign)? "" : "-", key);
#endif
    }

    if (sign) {
      free_trie(no->pos);
      no->pos = CLAUSE3;
    
      if (no->neg != NULL) {
	no->bro = merge_tries(no->bro, no->neg);
	no->neg = NULL;
      }
    } else {
      free_trie(no->neg);
      no->neg = CLAUSE3;    
    
      if (no->pos != NULL) {
	no->bro = merge_tries(no->bro, no->pos);
	no->pos = NULL;
      }
    }
    
    if (no->bro != CLAUSE3) return root;
  }
  return CLAUSE3;
}

int trie_subsume_clause (no, arr, signs, count)
     struct trie *no;
     int arr[], signs[]; /* current clause */
     int count; /* the length of the current clause */
{
  int key;
  while (no != NULL) {
    if (no == CLAUSE3) return 1;
    if (count > 0) key = arr[--count]; else return 0;
    if (key == no->key) {
      if (signs[key] == 1) no = no->pos;
      else no = no->neg;
    } else if (key > no->key) {
      no = no->bro;
      while ((no != NULL) && (key > no->key)) {
	no = no->bro;
      }
    }
  }
  return 0;
} /* trie_subsume_clause */


struct trie *merge_tries(n1, n2)
     struct trie *n1, *n2;  
     /* do lazy clean and complete merge */
{
    if (n2 == NULL) return n1;  
    if (n2 == CLAUSE3) { free_trie(n1); return n2;}
    if (n1 == NULL) return n2;
    if (n1 == CLAUSE3) { free_trie(n2); return n1;}
    
    if (n2->key < n1->key) {
    n2->bro = merge_tries(n1, n2->bro);
    if (n2->bro == CLAUSE3) { free_trie(n2); return CLAUSE3; }
    return n2;
  }

  if (n1->key < n2->key) {
    n1->bro = merge_tries(n1->bro, n2);
    if (n1->bro == CLAUSE3) { free_trie(n1); return CLAUSE3; }
    return n1;
  }

  /* now, n1->key == n2->key */
  n1->bro = merge_tries(n1->bro, n2->bro);
  n2->bro = NULL;
  if (n1->bro == CLAUSE3) {
    free_trie(n1);
    free_trie(n2);
    return CLAUSE3;
  }

  n1->pos = merge_tries(n1->pos, n2->pos);
  n1->neg = merge_tries(n1->neg, n2->neg);
  free_trie_cell(n2);

  if (n1->pos == CLAUSE3) {
    n1->bro = merge_tries(n1->bro, n1->neg);
    n1->neg = NULL;
  }
  if (n1->neg == CLAUSE3) {
    n1->bro = merge_tries(n1->bro, n1->pos);
    n1->pos = NULL;
  }

  if (n1->bro == CLAUSE3) { free_trie(n1); return CLAUSE3; }
  return n1;
}

struct trie *get_trie ()
{
  struct trie *p;
  struct links *l;

  if (Trie_avail == NULL) {
    if (DATA == 3) {
      l = ( struct links * ) tp_alloc ( sizeof ( struct links ) );
    } else
      l = NULL;
    p = ( struct trie * ) tp_alloc ( sizeof ( struct trie ) );
    if (p == NULL) {
      fprintf ( stderr, "get_trie return NULL.\n");
      exit(1);
    }
    p->link = l;
    Trie_news++;
  } else {
    Trie_avails--;
    p = Trie_avail;
    Trie_avail = Trie_avail->bro;
  }

  Trie_gets++;
  p->pos = p->neg = p->bro = p->par = NULL;
  p->esign = DC;

  if (DATA == 3) {
    l = p->link;
    l->lin = l->back = NULL;
    l->stamp = NULL;
  }

  return(p);
} 
 
void free_trie (no)
     struct trie *no;
{
  struct trie *no2;
  while (no != NULL && no != CLAUSE3) {
    if (no->pos != NULL && no->pos != CLAUSE3) free_trie ( no->pos );
    if (no->neg != NULL && no->neg != CLAUSE3) free_trie ( no->neg );
    no2 = no;
    no = no->bro;
    free_trie_cell(no2);
  }
}

void print_trie (no)
     struct trie *no;
{
  /*
  printf("\nUnit Clauses: ");
  for (i = 1; i <= Max_atom; i++) 
    if (Value[i] != DC) {
      if (Value[i] == TT)
	fprintf(Sato_fp, "(%d) ", i);
      else
	fprintf(Sato_fp, "(-%d) ", i);
    }
    */
  fprintf(Sato_fp, "\nClause Tree:\n");
  if (no == NULL)
    fprintf(Sato_fp, "   nil");
  else if (no == CLAUSE3)
    fprintf(Sato_fp, "   [C]");
  else 
    print_trie_aux(no, 0, "", "", "");
  fprintf(Sato_fp, "\n");
}

void print_trie_aux(no, sign, leader, addition, arrow)
     /* assume no != NULL */
     struct trie *no;
     int sign;
     char *leader, *addition, *arrow;
{
  char leader2[MAX_ATOM];
  int i;
  i = 0;

  fprintf(Sato_fp, "\n%s%s", leader, arrow);

  if (sign == 1 || sign == 2)
    if (no == CLAUSE3)
      fprintf(Sato_fp, "->[C]");
    else 
      fprintf(Sato_fp, "-x%d[%d,%d]", no->key, no->esign, no->psign);
  else 
    fprintf(Sato_fp, " x%d[%d,%d]", no->key, no->esign, no->psign);

  if (no->pos != NULL) i++;
  if (no->neg != NULL) i++;
  if (no->bro != NULL) i++;

  str_cat(leader, addition, leader2);

  if (no->pos != NULL) {
    if (i == 1)
      print_trie_aux(no->pos, 1, leader2, "     ", " `-P-");
    else
      print_trie_aux(no->pos, 1, leader2, " |   ", " |-P-");
    i--;
  }

  if (no->neg != NULL) {
    if (i == 1)
      print_trie_aux(no->neg, 2, leader2, "     ", " `-N-");
    else
      print_trie_aux(no->neg, 2, leader2, " |   ", " |-N-");
    i--;
  }

  if (no->bro != NULL) {
      print_trie_aux(no->bro, 3, leader2, "", "");
    }
}

void print_clauses_in_trie ()
{
  int cl_arr[MAX_LIST], sign_arr[MAX_ATOM];

  if (Root == CLAUSE3) {
    print_clause(cl_arr, sign_arr, 0);
    fprintf(stderr, "\nAn empty clause is found.\n");
  } else if (Root != NULL) 
    print_cl_in_trie_aux(Root, 0, cl_arr, sign_arr);
}

void print_cl_in_trie_aux (no, size, cl_arr, sign_arr)
     /* assume no != NULL */
     struct trie *no;
     int size, cl_arr[], sign_arr[];
{
  if (no == CLAUSE3) {
    print_clause(cl_arr, sign_arr, size);
    return;
  } else {
    cl_arr[size] = no->key;
  }

  if (no->pos != NULL) {
    sign_arr[no->key] = TT;
    print_cl_in_trie_aux(no->pos, size+1, cl_arr, sign_arr);
  }

  if (no->neg != NULL) {
    sign_arr[no->key] = FF;
    print_cl_in_trie_aux(no->neg, size+1, cl_arr, sign_arr);
  }

  if (no->bro != NULL) {
    print_cl_in_trie_aux(no->bro, size, cl_arr, sign_arr);
  }
}

/***********************************************************/
/** search for a model **/
/***********************************************************/

trie_search ()
{
  int res = 0;

#ifdef MORETRACE
  if (TRACE == 6) print_trie(Root);
#endif

  if (Root == NULL) {
    Branch_num = Branch_succ = 1;
    fprintf(Sato_fp, "\nEvery atom has a fixed truth value.\n");
    if (TRACE > 0) print_model(Sato_fp);
    res = 1;
  } else if (Root == CLAUSE3 || trie_create_var_table() == UNSAT) {
    Branch_num = Branch_fail = 1;
    fprintf(Sato_fp, "\nAn empty clause is found from the first %d clauses.\n",
	   Clause_num);
    res = 0;
  } else {
    Branch_num = 1;
    Branch_fail = Branch_succ = 0;
    res = (DATA == 2)? trie2_search2() : trie_search2(); 
  }
  return res;
}

int trie_search2 ()
{
  int i;

  /* there is a BUG appearing in trie_push_brothers
     when the file BUG is input 
     */

  Max_cand = (RENAME>1)? RENAME : Max_atom+1;
  Tmp_idx = 0;
  i = (trie_push_brothers(Root) != UNSAT);

  /* the major loop goes here. */
  while (i && !Stop) {
    if (trie_propagate_values() != UNSAT)
      i = trie_next_key();
    else {
      i = trie_prev_key();
    }
  } /* while */

  return ((Branch_succ)? 1 : (Stop)? 2 : 0);
}

int trie_next_key ()
{
  int i;

  if (OLDV) {
    int flag = TT;

    while (OLDV && flag) {
      i = Old_value[--OLDV];
      if (i < 0) { i = -i; flag = FF; }

      if (Value[i] != DC) flag = TT;
      else {
	Top_var3 = V3[i];
	assign_value(i, flag); 

	if (Old_choice[OLDV] == TT) {
	  Backup[Backup_idx++] = i;
	  Branch_open++;
	  Branch_num++;
	} else {
	  Backup[Backup_idx++] = -i;
	}

#ifdef MORETRACE
	print_let_x_have;
	/*if (QGROUP && TRACE >= 2) print_cell(i);*/
#endif

	return i;
      }
    }
  }

  if (GREEDY <= Branch_open) {
    if (greedy_search()) {
      printf("Found a model by the Greedy Search.\n");
      return handle_succ_end(); 
    }
    return trie_prev_key();
  }

  if ((i = trie_best_key()) == 0) return handle_succ_end();

  Branch_open++;
  Branch_num++;
  
  Top_var3 = V3[i];
  if (SELECT == 6) {
    int sign = rand() % 2;
    assign_value(i, sign);
  } else {
    assign_value(i, CHOICE1); 
  }

#ifdef MORETRACE
  print_let_x_have;
  /*if (QGROUP && TRACE >= 2) print_cell(i);*/
#endif

  if (Backup_idx >= MAX_LIST) {
    fprintf ( stderr, "MAX_LIST(%d) is too small in next_key(%d).\n", 
	      MAX_LIST, i);
    Stop = 1;
  }
  Backup[Backup_idx++] = i;

  return i;
}

int trie_best_key()
{
  static int first = -1;
  if (first == -1 && JUMP > 1) { first = JUMP; }
  if (first > 0) { first--;  return strategy0();}

  if (SELECT == 0) { 
    return trie_strategy2();
  } else if (SELECT == 10) {
    return strategy0();
  } else if (SELECT == 1) {
    return trie_strategy1();
  } else if (SELECT == 2) { 
    return trie_strategy2();
  } else if (SELECT == 3) {
    return trie_strategy3();
  } else if (SELECT == 4) {
    return trie_strategy4();
  } else if (SELECT == 5) {
    return trie_strategy5();
  } else if (SELECT > 5 && SELECT < 10) {
    return trie_strategy2();
  } else {
    fprintf ( stderr, "Unknown option: -s%d\n", SELECT);
    return (Stop = 1);
  }
}

int trie_list_len (no)
     struct trie *no;
{ 
  int i = 0;
  while (no != CLAUSE3) { no = no->link->lin; i++; }
  return i;
}

int trie_strategy1 ()
     /* the least unsigned variable in a shortest positive clause */
{
  struct trie *no;
  int i, first_i, best_i, best;
  unsigned int k;
  best = Max_atom;
  first_i = best_i = 0;

  for (i = 1; i < Max_cand; i++) 
    if (Value[i] == DC) {
      if (first_i == 0) first_i = i;
      no = V3[i]->both->tr;
      while (no != CLAUSE3) {
	if ((k = trie_short_pos_clause_all(no, best)) < best) 
	  if (k == 1) {
	    return i;
	  } else {
	    best = k;
	    best_i = i;
	  }
	no = no->link->lin;
      }
      no = V3[i]->poss->tr;
      while (no != NULL) {
	if ((k = trie_short_pos_clause_all(no, best)) < best) 
	  if (k == 1) {
	    return i;
	  } else {
	    best = k;
	    best_i = i;
	  }
	no = no->link->lin;
      }
    }
  return (best_i == 0)? first_i : best_i;
}

int trie_strategy2 ()
     /* the least unsigned variable in a shortest 
	non-horn clause */
{
  int key, i;
  int min_key = 0;
  int min = MAX_ATOM;
  struct trie *cl = NHtrie;
  struct trie *son, *no, *min_clause;

  while (cl != NULL) {
    i = 0;
    key = 0;
    son = cl;
    no = son->par;
    while (no != NULL) {
      if (Value[no->key] == DC) {
	if (++i >= min) goto end; 
	if (no->key < Max_cand) {
	  key = no->key;
	}
      } else if (Value[no->key] == son->psign) 
	goto end;

      son = no; 
      no = son->par;
    }	

    if (key) {
      if (i <= 2) { 
#ifdef MORETRACE
	if (TRACE == 4) print_clause_from_leaf(cl);
#endif
	return key;
      }
      min = i;
      min_key = key;
      min_clause = cl;
    }

  end:
    cl = cl->bro;
  }

#ifdef MORETRACE
  if (min_key && TRACE == 4) print_clause_from_leaf(min_clause);
#endif

  if (min_key || RENAME==0) return min_key;
  else return strategy0();
}

int trie_strategy3 ()
     /* the least unsigned variable in a shortest initially 
	non-horn clause */
{
  int key, i, j;
  int tmp_keys[200], best_keys[200], best_occurs[200];
  int best_num = 0;
  int min = MAX_ATOM-1;
  struct trie *son, *no;
  struct trie *cl = NHtrie;

  while (cl != NULL) {
    son = cl;
    no = son->par;
    i = key = 0;
    while (i < min && no != NULL) {
      if (Value[no->key] == DC) {
	tmp_keys[i++] = key = no->key;
	son = no; 
	no = son->par;
      } else if (Value[no->key] == son->psign) {
	i = MAX_ATOM;
      } else {
	son = no; 
	no = son->par;
      }
    }	

    if (min > i) {
      best_num = min = i;
      for (i = 0; i < min; i++) {
	best_keys[i] = tmp_keys[i];
	best_occurs[i] = 1;
      }
    } else if (min == i) {
      for (i = 0; i < min; i++) {
	key = tmp_keys[i];
	for (j = 0; j < best_num; j++) 
	  if (best_keys[j] == key) { best_occurs[j]++; j = MAX_ATOM; }
	if (j < MAX_ATOM) {
	  best_keys[best_num] = key;
	  best_occurs[best_num++] = 1;
	}
      }
    }
    cl = cl->bro;
  }

  if (!best_num) return 0;

  j = 0;
  for (i = 1; i < best_num; i++) {
    if ((best_occurs[i] > best_occurs[j])) {
      j = i;
    }
  }

  return best_keys[j];
}

int trie_strategy4 ()
     /* the least unsigned variable in a shortest 
	non-horn clause */
{
  int key, i, j;
  int tmp_keys[200], best_keys[MAX_ATOM], best_occurs[200];
  int best_num = 0;
  int min = MAX_ATOM-1;
  struct trie *son, *no;
  struct trie *cl = NHtrie;

  while (cl != NULL) {
    son = cl;
    no = son->par;
    i = 0;
    while (i < min && no != NULL) {
      if (Value[no->key] == DC) {
	tmp_keys[i++] = key = no->key;
	son = no; 
	no = son->par;
      } else if (Value[no->key] == son->psign) {
	i = MAX_ATOM;
      } else {
	son = no; 
	no = son->par;
      }
    }	

    if (min > i) {
      best_num = min = i;
      for (i = 0; i < min; i++) {
	key = best_keys[i] = tmp_keys[i];
      }
    } else if (min == i) {
      for (i = 0; i < min; i++) {
	key = tmp_keys[i];
	for (j = 0; j < best_num && (best_keys[j] != key); j++);
	if (j >= min) {
	  best_keys[best_num++] = key;
	}
      }
    }
    cl = cl->bro;
  }

  if (!best_num) return 0;

  for (i = 0; i < best_num; i++) {
    struct var3 *el = V3[best_keys[i]];
    best_occurs[i] = 
      trie_list_len(el->both->tr) +
      trie_list_len(el->pnegs->tr) +
      trie_list_len(el->pposs->tr) +
      trie_list_len(el->negs->tr) +
      trie_list_len(el->poss->tr);
  }
  j = 0;
  for (i = 1; i < best_num; i++) {
    if ((best_occurs[i] > best_occurs[j])) {
/*      printf("length = %d, key = %d, best occ = %d\n",
	     min, best_keys[j], best_occurs[j]);
*/
      j = i;
    }
  }

/*  printf("LENGTH = %d, key = %d, best occ = %d\n",
	 min, best_keys[j], best_occurs[j]);
*/
  return best_keys[j];
}

int trie_strategy5 ()
     /* the least unsigned variable in a shortest 
	non-horn clause */
{
  int key, i, j;
  int tmp[500], best_keys[500];
  int best_num = 0;
  int min = MAX_ATOM-1;
  struct trie *son, *no;
  struct trie *cl = NHtrie;

  while (cl != NULL) {
    i = 0;
    son = cl;
    no = son->par;
    while (no != NULL) {
      if (Value[no->key] == DC) {
	if (++i > min) goto end; 
	tmp[i] = key = no->key;
      } else if (Value[no->key] == son->psign) 
	goto end;

      son = no; 
      no = son->par;
    }	

    if (i < 3 && Lookahead_level >= Backup_idx) 
      return tmp[1];
    else if (min > i) {
      best_num = min = i;
      for (i = 0; i < min; i++) {
	best_keys[i] = tmp[i+1];
      }
    } else { /* (min == i) */
      for (i = 1; i <= min; i++) {
	key = tmp[i];
	for (j = 0; j < best_num; j++) 
	  if (best_keys[j] == key) goto exist;
	best_keys[best_num++] = key;
      exist: ;
      }
    }

  end:
    cl = cl->bro;
  }

  if (!best_num) return 0;

#ifdef RESEARCH
  if (Lookahead_level < Backup_idx) {
    for (i = 0; i < best_num; i++) tmp[i] = TT;
    if (trie_look_ahead(best_num, best_keys, tmp) == UNSAT)
      return UNSAT;
    best_key = best_val = 0;
    for (i = 0; i < best_num; i++) {
      key = best_keys[i];
      if ((Value[key] == DC) && (tmp[i] > best_val)) {
	best_key = key;
	best_val = tmp[i];
      }
    }
    return best_key;
  } else
#endif
    return best_keys[0];
}

int trie_prev_key ()
{
  int i, j;
  int jump = JUMP;

  OLDV = Tmp_idx = 0;
  j = Branch_fail & two15;

  while (Backup_idx-- > 0) {
    if ((i = Backup[Backup_idx]) < 0) {
      trie_clean_dependents(V3[-i]);
      --jump;
    } else {
      int sign = Value[i];
      trie_clean_dependents(V3[i]);
      Branch_open--;

      if (j || --jump < 1) {
	Backup[Backup_idx++] *= -1;
	Top_var3 = V3[i];
	assign_value(i, NEG(sign));

#ifdef MORETRACE
	print_now_let_x_have;
#endif
	return i;
      } else
	Branch_skip++;
    }
  }

  if (GREEDY < MAX_ATOM) printf("\nThe greedy search for models has failed.\n");
  if (JUMP > 1) printf("\nThe jumpy search for models has failed.\n");

  return Stop = 0;
}

#ifdef RESEARCH
int trie_look_ahead (total, p_arr, sign_arr)
     int total, p_arr[], sign_arr[];
     /* return UNSAT if it results in inconsistency
	when P[i] is assigned SIGN[i] */
{
  struct var3 *old_top_var;
  int res, i, p, p_sign;

  for (i = 0; i < total; i++) if (Value[p = p_arr[i]] == DC) {
    old_top_var = Top_var3;
    Top_var3 = V3[p];

    assign_value(p, p_sign = sign_arr[i]);

#ifdef MORETRACE
    if (TRACE > 3) printf("     Looking ahead on %s%d\n", 
			  (p_sign == TT)? "x" : "-x", p);
#endif

    res = trie_propagate_values();
    if (res == UNSAT) {
      Tmp_idx = 0;
      trie_clean_dependents(V3[p]);
      Top_var3 = old_top_var;

      trie_assign_value(p, NEG(p_sign));

#ifdef MORETRACE
      if (TRACE > 3) printf("        %s%d is set to false by lookahead\n", 
			    (p_sign == TT)? "x" : "-x", p);
#endif

      if (trie_propagate_values() == UNSAT) {
	return UNSAT;
      }
    } else {
      int j;
      res = 0;
      for (j = 1; j <= Max_atom; j++) if (Value[j] != DC) res++;
      sign_arr[i] = res;

      trie_clean_dependents(V3[p]);
      Top_var3 = old_top_var;
    }
  }
  return SAT;
}
#endif

int trie_propagate_values ()
{
  struct var3 *el;
  int i;

  while (Tmp_idx > 0) {
    el = V3[i = Tmp[--Tmp_idx]];
    if (Value[i] == TT) {
      if (trie_push_list_neg(el->both->tr) == UNSAT) return UNSAT;
      if (trie_push_list_neg(el->negs->tr) == UNSAT) return UNSAT;
      if (trie_pull_list(el->pnegs->tr) == UNSAT) return UNSAT;
    } else if (Value[i] == FF) {
      if (trie_push_list_pos(el->both->tr) == UNSAT) return UNSAT;
      if (trie_push_list_pos(el->poss->tr) == UNSAT) return UNSAT;
      if (trie_pull_list(el->pposs->tr) == UNSAT) return UNSAT;
    } else {
      printf("value[x%d] = %d\n", i, Value[i]);
    }
  }
  return SAT;
}

int trie_pull_list (top_no)
     struct trie *top_no;
{
  struct trie *no;
  int i, sign;

  while (top_no != CLAUSE3) {
    no = top_no->par;
    sign = top_no->psign;

    /*if (trie_pull_par_up(no->par, no->psign) == UNSAT) return UNSAT; */

    while (1) {
      while (Value[no->key] != DC) {
	if (Value[no->key] == sign) { goto end; } 
	else {
	  sign = no->psign;
	  if ((no = no->par) == NULL) return handle_fail_end();
	}
      }

      if (no->esign == DC) {
	if (no->link->lin != NULL) {
	  i = no->key;
	  trie_assign_value(i, sign);
#ifdef MORETRACE
	  print_x_has_to;
#endif

	} else {
	  struct cap *cp;
	  cp = (no->esign = sign)? V3[no->key]->pposs : V3[no->key]->pnegs;
	  no->link->lin = cp->tr;
	  cp->tr = no;
	  if (Backup_idx) {
	    no->link->stamp = cp;
	    no->link->back = Top_var3->pull;
	    Top_var3->pull = no;
	  }
	}
	goto end;
      } else if (no->esign == sign) {
	goto end;
      } else {
	sign = no->psign;
	if ((no = no->par) == NULL) return handle_fail_end();
      }
    }
  end: 
    top_no = top_no->link->lin;
  }    
  return SAT;
}

int trie_pull_par_up (no, sign)
     struct trie *no;
     int sign;
{
  int i;

  while (1) {
    while (Value[no->key] != DC) {
      if (Value[no->key] == sign) return SAT;
      sign = no->psign;
      if ((no = no->par) == NULL) return handle_fail_end();
    }
   
    if (no->esign == DC) {
      if (no->link->lin != NULL) {
	i = no->key;
	trie_assign_value(i, sign);
#ifdef MORETRACE
	print_x_has_to;
#endif

      } else {
	struct cap *cp;
	cp = (no->esign = sign)? V3[no->key]->pposs : V3[no->key]->pnegs;
	no->link->lin = cp->tr;
	cp->tr = no;
	if (Backup_idx) {
	  no->link->stamp = cp;
	  no->link->back = Top_var3->pull;
	  Top_var3->pull = no;
	}
      }
      return SAT;
    } else if (no->esign == sign) {
      return SAT;
    } else {
      sign = no->psign;
      if ((no = no->par) == NULL) return handle_fail_end();
    }
  }
}

int trie_push_list_pos (top_cl)
     struct trie *top_cl;
{
  while (top_cl != CLAUSE3) {
#ifdef MORETRACE
    if (trie_push_brothers(top_cl->pos) == UNSAT) return UNSAT;
#else
    struct trie *tr = top_cl->pos;
    TRIE_PUSH_BROTHERS 
#endif
    top_cl = top_cl->link->lin;
  }    
  return SAT;
}

int trie_push_list_neg (top_cl)
     struct trie *top_cl;
{
  while (top_cl != CLAUSE3) {
#ifdef MORETRACE
    if (trie_push_brothers(top_cl->neg) == UNSAT) return UNSAT;
#else
    struct trie *tr = top_cl->neg;
    TRIE_PUSH_BROTHERS
#endif
    top_cl = top_cl->link->lin;
  }    
  return SAT;
}

int trie_push_brothers (tr)
     struct trie *tr;
{
  struct trie *bros[MAX_SHORT];
  struct trie *no;
  int i;
  int idx;
  struct var3 *it;


  idx = 1;
  bros[0] = tr;
  while (idx > 0) {
    tr = bros[--idx];
    while (tr != NULL) {
      no = tr;
      i = no->key;
      it = V3[i];

      if (no->esign == TT) {

	if (Value[i] == DC) {
	  trie_assign_value(i, TT);
#ifdef MORETRACE
	  print_x_has_to;
#endif
	} else if (Value[i] == FF) {
#ifdef MORETRACE
	  print_x_contradictory;
#endif
	  return handle_fail_end();
	}

	if (tr->bro != NULL) bros[idx++] = tr->bro;
	tr = no->neg;

      } else if (no->esign == FF) {

	if (Value[i] == DC) {
	  trie_assign_value(i, FF);
#ifdef MORETRACE
	  print_x_has_to;
#endif
	} else if (Value[i] == TT) {
#ifdef MORETRACE
	  print_x_contradictory;
#endif
	  return handle_fail_end();
	}

	if (tr->bro != NULL) bros[idx++] = tr->bro;
	tr = no->pos;

      } else { /* no->esign == DC */

	if (Value[i] == TT) {
	  if (tr->bro != NULL) bros[idx++] = tr->bro;
	  tr = no->neg;
	} else if (Value[i] == FF) {
	  if (tr->bro != NULL) bros[idx++] = tr->bro;
	  tr = no->pos;
	} else {  /* Value[i] == DC */ 
	  struct cap *cp;
	  if (no->neg == NULL) {
	    if (no->pos != NULL) {
	      cp = it->poss;
	      no->link->lin = cp->tr;
	      cp->tr = no;
	      if (Backup_idx) {
		no->link->stamp = cp;
		no->link->back = Top_var3->push;
		Top_var3->push = no;
	      }
	    }
	  } else if (no->pos == NULL) {
	    cp = it->negs;
	    no->link->lin = cp->tr;
	    cp->tr = no;
	    if (Backup_idx) {
	      no->link->stamp = cp;
	      no->link->back = Top_var3->push;
	      Top_var3->push = no;
	    }
	  } else {
	    cp = it->both;
	    no->link->lin = cp->tr;
	    cp->tr = no;
	    if (Backup_idx) {
	      no->link->stamp = cp;
	      no->link->back = Top_var3->push;
	      Top_var3->push = no;
	    }
	  }
	  tr = tr->bro;
	}
      }
    }
  }
  return SAT;
}

void trie_clean_dependents (top)
     struct var3 *top;
{
  int i;
  struct trie *no;
  struct var3 *it, *next;

  it = top;
  while (it != NULL) {
    i = it->key;
#ifdef MORETRACE
    print_up_to_x;
#endif
    Value[i] = DC;
    next = it->next;
    it->next = NULL;
    it = next;
  }

  no = top->pull;
  top->pull = NULL;
  while (no != NULL) {
    no->link->stamp->tr = no->link->lin;
    no->link->lin = NULL;
    no->esign = DC;
    no = no->link->back;
  }

  no = top->push;
  top->push = NULL;
  while (no != NULL) {
    no->link->stamp->tr = no->link->lin;
    no->link->lin = NULL;
    no = no->link->back;
  }
}

int trie_short_pos_clause(no, min)
     struct trie *no;
     int min;
{
  int i = 1;

  while (no != NULL) {
    if (no->esign == TT) return i;
    if (no->esign == FF) return min;
    if (Value[no->key] == TT) 
      no = no->neg;
    else {
      if (Value[no->key] == DC && (++i >= min)) return min;
      no = no->pos;
    }
  }
    
  return min;
}

int trie_short_pos_clause_all(no0, min)
     /* count the length of the shortest positive clause in no0. */
     struct trie *no0;
     int min;
{
  struct trie *no;
  int key = no0->key;

  if (min == 1) return 1;
  if (Value[key] == DC) {

    no = no0->pos;
    if (no != NULL) {
      if (no->esign == TT) return 1;
      if (no->esign == FF) return min;
      while (no != NULL && no->pos != NULL) {
	min = trie_short_pos_clause_all(no, min);
	if (min == 1) return 2;
	no = no->bro;
      }
    }
    return ++min;

  } else {

    if (Value[key] == FF) 
      no = no0->pos;
    else no = no0->neg;

    if (no != NULL) {
      if (no->esign == TT) return 1;
      if (no->esign == FF) return min;
      while (no != NULL && no->pos != NULL) {
	min = trie_short_pos_clause_all(no, min);
	if (min == 1) return 1;
	no = no->bro;
      }
    }
    return min;
  }
}

void pv(n) int n; { printf("Value[%d] = %d", n, Value[n]); }


void p_trie_info (f)
     FILE *f;
{
  fprintf(f, "\nThere are %d nodes in use and %d nodes available.\n",
	 Trie_gets - Trie_frees, Trie_avails);
  fprintf(f, "There are %d calls to get_node and %d calls to new_node.\n",
	 Trie_gets, Trie_news);
}

void init_var3_table ()
{
  static int old_Max_atom;
  int i;
  struct var3 *el;

  if (V3 != NULL) {
    for (i=1; i <= old_Max_atom; i++)  {
      el = V3[i];
      free(el->poss);
      free(el->negs);
      free(el->both);
      free(el->pnegs);
      free(el->pposs);
    }
    free(V3);
  }

  old_Max_atom = Max_atom;
  V3 = (struct var3 **) malloc ( sizeof ( struct var3 *) * (1+Max_atom ));
  Memory_count += ( sizeof ( struct var3 *) * (1+Max_atom ));

  if (V3 == NULL) {
    fprintf ( stderr, "var_table returns NULL.\n");
    exit(1);
  }

  for (i=0; i <= Max_atom; i++)  {
    V3[i] = el = (struct var3 *) tp_alloc ( sizeof ( struct var3 ) );

    el->poss = (struct cap *) tp_alloc( sizeof ( struct cap ));
    el->negs = (struct cap *) tp_alloc( sizeof ( struct cap ));
    el->both = (struct cap *) tp_alloc( sizeof ( struct cap ));
    el->pnegs = (struct cap *) tp_alloc( sizeof ( struct cap ));
    el->pposs = (struct cap *) tp_alloc( sizeof ( struct cap ));

    el->both->tr = el->poss->tr = el->negs->tr = CLAUSE3;
    el->pposs->tr = el->pnegs->tr = CLAUSE3;
    el->push = NULL;
    el->pull = NULL;
    el->key = i;
    el->next = NULL;
  }
}

int visit_trie (tr, par, pflag, pos)
     struct trie *tr, *par;
     int pflag, pos;
{
  if (tr == CLAUSE3) {
    if (trie_pull_par_up(par, pflag) == UNSAT) {
      /* free_trie(Root);*/
      Root = CLAUSE3; 
      return 0; 
    }
    if (pos > 1) {
      tr = get_trie();
      tr->key = 0;
      tr->par = par;
      tr->psign = pflag;
      tr->bro = NHtrie;
      NHtrie = tr;
      NH_num++;
    }
    return 1;
  } else if (Stop) {
    return 0;
  } else {
    int hascl = 0;
    while (tr != NULL) {
      tr->par = par;
      tr->psign = pflag;
      tr->esign = DC;
      if (tr->pos != NULL && visit_trie(tr->pos, tr, 1, pos+1)) 
	hascl = 1;
      if (tr->neg != NULL && visit_trie(tr->neg, tr, 0, pos))
	hascl = 1;
      tr = tr->bro;
    }
    return hascl;
  } 
}

/********************************************************
  Toplevel Code for a different search algorithm: DATA = 2
********************************************************/

int trie_create_var_table()
{
  int pos, neg, i;
  struct var2 *el;
  int extra_space = 0;
  if (GROW) extra_space = EXTRA_SPACE;

  if (Root == CLAUSE3) return UNSAT;
  if (Root == NULL) return SAT;

  if (DATA == 3) {
    init_var3_table ();
    visit_trie(Root, (struct trie *) NULL, DC, 0);
    if (Root == CLAUSE3) return UNSAT;

  } else {

    init_var2_table ();
    Clause_num = visit_trie2(Root, NULL, 0, 0, 0);
    printf("There are %ld clauses in the trie (%d are non-Horn).\n", 
	   Clause_num, NH_num);
    if (SELECT == 0) trie2_record_keeping();

    for (i = 1; i <= Max_atom; i++) {
      el = V2[i];
      pos = el->pos_count;
      neg = el->neg_count;
      el->pos_count = el->neg_count = 0;

      if (Value[i] == DC) {
	if (pos == 0 || neg == 0) {
	  if (pos == 0) {
	    trie2_assign_value(i, FF);
	  } else { 
	    trie2_assign_value(i, TT);
	  }
	  Pure_num++;

#ifdef MORETRACE
	  print_x_pure;
#endif
	} 
      }

      if (Value[i] == DC) {
	el->pos_extra = el->neg_extra = extra_space;
	el->pos = get_trie2_array(pos+extra_space);
	el->neg = get_trie2_array(neg+extra_space);
      } else {
	level_of(el) = 0;
	if (pos > 0) el->pos = get_trie2_array(pos);
	if (neg > 0) el->neg = get_trie2_array(neg);
      }
    }

    trie2_fill_in_clauses(NHtrie);
    trie2_fill_in_clauses(CLtrie);
  }

  return SAT;
}

void init_var2_table ()
{
  static int old_Max_atom;
  int i;
  struct var2 *el;

  if (V2 != NULL) {
    for (i=1; i <= old_Max_atom; i++)  {
      el = V2[i];
      free(el->pos);
      free(el->neg);
    }
    free(V2);
  }

  old_Max_atom = Max_atom;
  V2 = (struct var2 **) malloc ( sizeof ( struct var2 *) * (1+Max_atom ));
  Memory_count += ( sizeof ( struct var2 *) * (1+Max_atom ));

  if (V2 == NULL) {
    fprintf ( stderr, "var_table returns NULL.\n");
    exit(1);
  }

  for (i=0; i <= Max_atom; i++)  {
    V2[i] = el = (struct var2 *) tp_alloc ( sizeof ( struct var2 ) );

    /* basic stuff */
    el->key = i;
    el->next = NULL;
    el->pos = el->neg = NULL;
    el->neg_count = el->pos_count = 0;

    /* related to unit-propagations. */
    el->frozen = NULL;                  /* frozen clauses */
    el->pure = 0; 

    /* GROW techniques */
    el->force = NULL;
    el->level = 0;
    el->pos_extra = el->neg_extra = 0;
  }

  /*check_cl(0, NULL, NULL); */
  Trie2_Ucl = get_trie2_array(MAX_LIST);
  Trie2_Ucl_idx = 0;
}

/********************************************************
  Code for a different search algorithm: DATA = 2
********************************************************/

struct trie **get_trie2_array (size)
     int size;
{
  struct trie **p;

  p = (struct trie **) malloc ( sizeof ( struct trie *) * size );
  if (p == NULL) {
    fprintf ( stderr, "get_trie2_array return NULL.\n");
    exit(1);
  }
  Memory_count += sizeof ( struct trie *) * size;
  return p;
}

long visit_trie2 (tr, par, pflag, count, pcount)
     struct trie *tr, *par;
     int count, pcount;
     /* Visit the whole trie.
	When go down, (a) assign the parent; (b) count the numbers
	of literals and positive literals; (d) if a CLAUSE3 is found, 
	create a clause node.
	When go up, return the number of clauses in the trie.
	*/
{
  if (tr == CLAUSE3) {
    tr = get_trie();
    tr->par = par;
    count_of_clause(tr) = count;
    tr->psign = pcount;

    if (pcount > 1) {
      tr->bro = NHtrie;
      NHtrie = tr;
      NH_num++;
    } else {
      tr->bro = CLtrie;
      CLtrie = tr;
    }

    return 1;

  } else {

    long x, y, z;
    int key = tr->key;
    x = y = z = 0;

    if (tr->pos != NULL) {
      y = visit_trie2(tr->pos, tr, 1, 1+count, 1+pcount);
      V2[key]->pos_count += y;
    }

    if (tr->neg != NULL) {
      z = visit_trie2(tr->neg, tr, 0, 1+count, pcount);
      V2[key]->neg_count += z;
    }

    if (tr->bro != NULL) 
      x = visit_trie2(tr->bro, par, pflag, count, pcount);

    tr->par = par;
    tr->psign = pflag;

    return (x + y + z);
  }
}

void trie2_fill_in_clauses (cl)
     struct trie *cl;
{
  struct trie *tr, *son;
  struct var2 *el;

  while (cl != NULL) {
    tr = cl->par;
    el = V2[tr->key];
    if (HAS_POS_CL(tr)) {
      el->pos[el->pos_count++] = cl;
    } else {
      el->neg[el->neg_count++] = cl;
    }

    son = tr;
    tr = tr->par;
    while (tr != NULL) {
      el = V2[tr->key];
      if (son->psign) {
	el->pos[el->pos_count++] = cl;
      } else {
	el->neg[el->neg_count++] = cl;
      }
      son = tr;
      tr = tr->par;
    }

    cl = cl->bro;
  }
}

void var_clauses( i, f )
     int i;
     /* f = 1: positive clauses only
	f = 2: negative clauses only
	f = 3: all clauses */
{
  int j, l;

  if (f != 2 && (l = V2[i]->pos_count) > 0) {
    printf("Here are the positive clauses for var %d.\n", i);
    for (j=0; j<l; j++) 
      print_clause_from_leaf(V2[i]->pos[j]);
  }

  if (f > 1 && (l = V2[i]->neg_count) > 0) {
    printf("Here are the negative clauses for var %d.\n", i);
    for (j=0; j<l; j++) 
      print_clause_from_leaf(V2[i]->neg[j]);
  }
}

void print_clause_from_leaf (tr)
     struct trie *tr;
{
  struct trie *son;

  printf("%s c=%d ( ", 
	 (active_clause(tr))? " " : "m", count_of_clause(tr));

  tr = tr->par;
  printf("%s%d[%d] ",  HAS_POS_CL(tr)? "" : "-", tr->key, Value[tr->key]);

  son = tr;
  tr = tr->par;
  while (tr != NULL) {
    printf("%s%d[%d] ",  (son->psign)? "" : "-", tr->key, Value[tr->key]);
    son = tr; 
    tr = tr->par;
  }
  printf(")\n");
}

int trie2_search2 ()
{
  struct var2 *it;
  int i;    
  
  Max_cand = (RENAME>1)? RENAME : Max_atom+1;
  if (trie2_handle_known_values() == UNSAT) return 0;

  Nonunit_cls = NHtrie;
  i = trie2_next_key();
  
  /* the major loop goes here. */
  while (!Stop && i > 0) {
    Top_var2 = it = V2[i];

    if (Value[i] == TT) {
      if (trie2_propagate_true(it,0) != UNSAT)
	i = trie2_next_key();
      else {
	i = trie2_prev_key(0);
      }
    } else { /* Value[i] == FF */
      if (trie2_propagate_false(it,0) != UNSAT) {
	i = trie2_next_key();
      } else {
	i = trie2_prev_key(0);
      }
    }
  } /* while */

  return ((Branch_succ > 0)? 1 : (Stop)? 2 : 0);
}

int trie2_next_key ()
{
  int i;
  
  if (trie2_handle_unit_clauses(1) == UNSAT)
    return trie2_prev_key(1);
  
  if (OLDV) {
    int flag = TT;

    while (OLDV && flag) {
      i = Old_value[--OLDV];
      if (i < 0) { i = -i; flag = FF; }

      if (Value[i] != DC) flag = TT;
      else {
	Top_var2 = V2[i];
	Value[i] = flag;
	level_of(Top_var2) = Branch_open;
	mark_of(Top_var2) = NULL;

	if (Old_choice[OLDV] == TT) {
	  Branch_open++;
	  Branch_num++;
	  Backup[Backup_idx++] = i;
	} else
	  Backup[Backup_idx++] = -i;

#ifdef MORETRACE
	print_let_x_have;
#endif

	return i;
      }
    }
  }

  UIP = HAVE_UIP;
  i = trie2_best_key();

  if (i == 0) return handle_succ_end();

  Branch_open++;
  Branch_num++;
  
  if (Backup_idx >= MAX_LIST) {
    fprintf ( stderr, "MAX_LIST(%d) is too small in trie2_next_key().\n", 
	      MAX_LIST);
    Stop = 1;
  }

  Backup[Backup_idx++] = i;
  level_of(V2[i]) = Branch_open;
  mark_of(V2[i]) = NULL;

#ifdef MORETRACE
  print_let_x_have;
#endif

  return i;
}

int trie2_prev_key (clean)
     int clean;
{
  int i;
  
  OLDV = Trie2_Ucl_idx = Tmp_idx = 0;
  Tmp_idx_end = MAX_ATOM;
  UIP = 0;
  
  while (Backup_idx-- > 0) {
    if ((i = Backup[Backup_idx]) < 0) {
      int j;
      
      i = -i; 
      Conflict_cl = NULL;
      j = Conflict_cl_level;

      if (clean) trie2_clean_dependents(V2[i], clean-1);
#ifdef MORETRACE
      else print_up_to_x;
#endif
      Value[i] = DC;
      clean = 2;

      if (GROW) {
	if (JUMP && j < Branch_open) {
	  int good = 1;
	  Jump_num++;

#ifdef MORETRACE
	  if (TRACE > 2)
	    printf("    Backjumping from level %d to %d\n", Branch_open, j);
#endif

	  while (good && Backup_idx) {
	    if ((i = Backup[--Backup_idx]) < 0) {
	      trie2_clean_dependents(V2[-i], clean);
	    } else if (level_of(V2[i]) > j) {
	      Branch_num--;
	      Branch_open--;
	      trie2_clean_dependents(V2[i], clean);
	    } else { 
	      Backup_idx++;
	      good = 0;
	    }
	  }
	}
      }

    } else {
      int NEXT_VALUE = NEG(Value[i]);

      Branch_open--;

      if (clean) trie2_clean_dependents(V2[i], clean-1);
#ifdef MORETRACE
      else print_up_to_x;
#endif

      Backup[Backup_idx++] = -i;
      Value[i] = NEXT_VALUE;

      if (GROW) {
	if (Conflict_cl == NULL) {
	  mark_of(V2[i]) = NULL;
	} else {
	  mark_of(V2[i]) = Conflict_cl;
	  Conflict_cl = NULL;
	}
	if (Conflict_cl_level > Branch_open) Conflict_cl_level = Branch_open;
	level_of(V2[i]) = Conflict_cl_level;
      }

#ifdef MORETRACE
      if (GROW && (TRACE > 8 || TRACE == 3 || TRACE == 4))
	  printf("  [%d]", level_of(V2[i]));
      print_now_let_x_have;
#endif
      return i;
    }
  }
  return Stop = 0;
}

int trie2_best_key()
{
  static int first = -1;
  if (first == -1 && JUMP > 1) { first = JUMP; }

  if (first > 0) {
    first--;
    return strategy0();
  }

  if (SELECT == 0) {
    return trie2_strategy2();
  } else if (SELECT == 1) {
    return trie2_strategy1();
  } else if (SELECT == 2) {
    return trie2_strategy2();
  } else if (SELECT == 3) {
    return trie2_strategy3();
  } else if (SELECT == 4) {
    return trie2_strategy4();
  } else if (SELECT == 5) {
    return trie2_strategy5();
  } else if (SELECT == 6 || SELECT == 7) {
    return trie2_strategy6();
  } else if (SELECT == 10) { 
    return strategy0();
  } else {
    fprintf ( stderr, "Unknown option: -s%d\n", SELECT);
    Stop = 1;
    return 0;
  }
}

int trie2_mom (total, keys)
     int total, keys[];
{
  int s, best_key, best_sign, sign, new, max_weight;
  best_sign = best_key = max_weight = 0;

#ifdef MORETRACE
  if (total && TRACE >= 4) 
    printf("There are %d MOM candidates.\n", total);
#endif

  for (s = 0; s < total; s++) {
    new = trie2_compute_mom(keys[s], &sign);
    if (new > max_weight) {
      best_key = keys[s];
      best_sign = sign;
      max_weight = new;
    }
  }

#ifdef MORETRACE
  if (total && TRACE >= 4) 
    printf("The winner is x%d.\n", best_key);
#endif

  Value[best_key] = best_sign;
  return best_key;
}

int trie2_compute_mom (i, sign)
     int i, *sign;
{
  int j, k, num_cls, pos_count, neg_count;
  struct trie *cl, **cls;

    pos_count = neg_count = 1;

    j = 0;
    cls = V2[i]->neg;
    num_cls = V2[i]->neg_count;
  
    for (k = 0; k < num_cls; k++) if (active_clause((cl=cls[k]))) { 
      j = 1; 
      if (count_of_clause(cl) == 2) {
	neg_count++;
      }
    }

    if (j == 0) {
      trie2_pure_delete(i, TT);
      return -1;
    }    

    j = 0;
    cls = V2[i]->pos;
    num_cls = V2[i]->pos_count;
  
    for (k = 0; k < num_cls; k++) if (active_clause((cl=cls[k]))) { 
      j = 1; 
      if (count_of_clause(cl) == 2) {
	pos_count++;
      }
    }
    
    if (j == 0) {
      trie2_pure_delete(i, FF);
      return -1;
    }    

#ifdef MORETRACE
  if (TRACE > 4) 
    printf("   key %d:  neg = %d, pos = %d\n", i, neg_count, pos_count);
#endif

  *sign = (neg_count > pos_count)? CHOICE2 : CHOICE1;
  return (neg_count * pos_count);
}

int trie2_strategy1 ()
{
  int i, total, keys[MAX_ATOM];

  total = 0;
  for (i = 1; i < Max_cand; i++)
    if (Value[i] == DC) keys[total++] = i;

  if (i = trie2_mom(total, keys)) return i;
  else {
    return strategy0();
  }
}

int trie2_strategy2 ()
     /* apply MOM to all active variables appearing
      in shortest non-Horn clauses */
{
  int i, x, y, min_len, total, keys[12];
  struct trie *cl = NHtrie;
  struct trie *cls[12];
  min_len = MAX_ATOM;
  total = 0;

  /* at first, collect all the shortnest clauses */
  while (cl != NULL) {
    if (active_clause(cl)) {
      if (count_of_clause(cl) < min_len) {
	min_len = count_of_clause(cl);
	cls[0] = cl;
	total = 1;
      } else if (total < MAGIC && count_of_clause(cl) == min_len) {
	cls[total++] = cl;
      }
    }
    cl = cl->bro;
  }

  if (total == 0) {
    return strategy0();
  }    

  /* then, colect active keys in the shortest clauses */
  i = total;
  y = 0;
  while  (i && y < MAGIC) {
    cl = cls[--i];
    cl = cl->par;
    while (cl != NULL && y < MAGIC) {
      x = cl->key;
      if (x < Max_cand && Value[x] == DC &&
	  V2[x]->pure != MAX_ATOM) {
	V2[x]->pure = MAX_ATOM;
	keys[y++] = x;
      }
      cl = cl->par;
    }
  }

  for (i = 0; i < y; i++) V2[keys[i]]->pure = 0;
  i = trie2_mom(y, keys);
  if (i == 0)  /* purecheck blocked the keys */
    return (trie2_strategy2());
  else 
    return i;
}

int trie2_strategy3 ()
     /* choose the min variable in the first shortest clause */
{
  int i;
  int min_clause_count = MAX_ATOM;
  struct trie *min_clause = NULL;
  struct trie *cl = NHtrie;

  while (cl != NULL) {
    if (active_clause(cl)) {
      i = count_of_clause(cl);
      if (i == cl->psign)
	if (min_clause_count > i) {
	  min_clause_count = i;
	  min_clause = cl;
	  if (i <= 2) cl = CLAUSE3;
	}
    }
    cl = cl->bro;
  }
  
  if (min_clause == NULL) return 0;

#ifdef MORETRACE
  if (TRACE == 4) print_clause_from_leaf(min_clause);
#endif

  i = 0;
  cl = min_clause->par;
  while (cl != NULL) {
    if (cl->key < Max_cand && Value[cl->key] == DC) i = cl->key;
    cl = cl->par;
  }
  if (i == 0) i = strategy0();
  Value[i] = CHOICE1;
  return i;
}

int trie2_strategy4 ()
  /* choose a variable in a shortest clause;
     using the occurrence as the second criterion */
{
  int i, min_idx;
  int min_clause_count = MAX_ATOM;
  struct trie *min_clauses[50], *no;
  struct trie *cl = NHtrie;

  for (i = 1; i <= Max_atom; i++) V2[i]->pos2 = 0;
  min_idx = 0;

  while (cl != NULL) {
    if (active_clause(cl)) {

      if (min_clause_count > i) {
	min_clause_count = i;
	min_idx = 1;
	min_clauses[0] = cl;
      } else if (min_idx < 50 && min_clause_count == i) {
	min_clauses[min_idx++] = cl;
      }

      no = cl->par;
      while (no != NULL) {
	V2[no->key]->pos2++;
	no = no->par;
      }
    }
    cl = cl->bro;
  }

  if (min_clause_count == MAX_ATOM) return 0;

  min_clause_count = 0;
  while (min_idx) {
    cl = min_clauses[--min_idx]->par;
    while (cl != NULL) {
      if (Value[cl->key] == DC && 
	  min_clause_count <  V2[cl->key]->pos2) {
	min_clause_count = V2[cl->key]->pos2;
	i = cl->key;
	/*printf("key = %d, occ = %d\n", i, min_clause_count); */
      }
      cl = cl->par;
    }
  }

  Value[i] = CHOICE1;
  return i;
}

int trie2_strategy5 ()
  /* choose a variable with the highest occurrence */
{
  int i, min_idx, min_key;
  struct trie *no;
  struct trie *cl = NHtrie;

  for (i = 1; i <= Max_atom; i++) V2[i]->pos2 = 0;

  while (cl != NULL) {
    if (active_clause(cl)) {
      no = cl->par;
      while (no != NULL) {
	V2[no->key]->pos2++;
	no = no->par;
      }
    }
    cl = cl->bro;
  }
  
  min_key = min_idx = 0;
  for (i = 1; i < Max_cand; i++) 
    if (Value[i] == DC && V2[i]->pos2 > min_idx) {
      min_idx = V2[i]->pos2;
      min_key = i;
    }

  Value[min_key] = CHOICE1;
  return min_key;
}

int trie2_strategy6 ()
  /* choose a variable with the highest occurrence */
{
  int i, min_wei, min_key, pos_wei;
  struct trie *no;
  struct trie *cl = NHtrie;

  for (i = 1; i <= Max_atom; i++) V2[i]->pos2 = 0;
  min_wei = 0;

  while (cl != NULL) {
    if (active_clause(cl)) {
      no = cl->par;
      while (no != NULL) {
	V2[no->key]->pos2 += 1;
	no = no->par;
      }
    }
    cl = cl->bro;
  }
  
  min_key = min_wei = 0;
  for (i = 1; i < Max_cand; i++) 
    if (Value[i] == DC && V2[i]->pos2 > min_wei) {
      min_wei = V2[i]->pos2;
      min_key = i;
    }

  if (min_key == 0) return 0;

  pos_wei = 0;
  for (i = V2[min_key]->pos_count-1; i >= 0; i--) 
    if (active_clause(V2[min_key]->pos[i])) pos_wei++;
  min_wei -= pos_wei;
  Value[min_key] = (min_wei >= pos_wei)? TT : FF; 
  return min_key;
}

void trie2_pure_delete (i, sign)
     int i, sign;
{
  int j, num_cls;
  struct trie **cls, *frozens, *cl;
  struct var2 *el;

  Value[i] = sign;
  el = V2[i];
  el->level = PURECHECK;
  mark_of(el) = NULL;
  Pure_num++;

#ifdef MORETRACE
  print_x_pure;
#endif

  el->next = Top_var2->next;
  Top_var2->next = el;

  if (sign == TT) {
    cls = el->pos;
    num_cls = el->pos_count;
  } else {
    cls = el->neg;
    num_cls = el->neg_count;
  }

  frozens = CLAUSE3;
  for (j = 0; j < num_cls; j++) if (active_clause((cl=cls[j]))) {
    frozen_lin(cl) = frozens;
    frozens = cl;
  }
  frozen_cls(el) = frozens;
}

void trie2_pure_check ()
     /* implementing pure deletion */
{
  int i, j, k, num_cls;
  struct trie **cls;
  struct var2 *el;

  for (i = 1; i <= Max_atom; i++) if (Value[i] == DC) {

    j = 0;
    el = V2[i];
    cls = el->neg;
    num_cls = el->neg_count;
    for (k = 0; k < num_cls; k++) 
      if (active_clause(cls[k])) { k = num_cls; j = 1; }

    if (j == 0) trie2_pure_delete(i, TT);
    else {

      j = 0;
      cls = el->pos;
      num_cls = el->pos_count;
      for (k = 0; k < num_cls; k++) 
	if (active_clause(cls[k])) { k = num_cls; j = 1; }

      if (j == 0) trie2_pure_delete(i, FF);
    }
  }
}

int trie2_propagate_true (el, save)
     struct var2 *el;
     int save;
{
  int key = el->key;
  int i, j, count, num_cls;
  struct trie **cls, *cl;

  /* decrement the count of clauses in which key appears negatively. */
  cls = el->neg;
  num_cls = el->neg_count;

  for (i = 0; i < num_cls; i++) {
    cl = cls[i];
    if (active_clause(cl)) {
      count = --count_of_clause(cl);
      if (count == 0) { 	
	/* restore what is done so far and then backtrack */
	count_of_clause(cl)++;

	for (j = 0; j < i; j++) {
	  if (active_clause(cls[j])) {
	    count_of_clause(cls[j])++;
	  }
	}
	if (save && GROW && Branch_open) trie2_record_conflict(cl);
	return handle_fail_end();
      } else if (count == 1) {
	/* store unit clauses in an array */
	Trie2_Ucl[Trie2_Ucl_idx++] = cl;
      }
    }
  }

  /* remember the key for freezing its positive clauses */
  Tmp[--Tmp_idx_end] = key;

  /* remember the vars for backtracking */
  if (save) {
    el->next = Top_var2->next;
    Top_var2->next = el;
  }
  return SAT;
}

int trie2_propagate_false (el, save)
     struct var2 *el;
     int save;
{
  int key = el->key;
  int i, j, count, num_cls;
  struct trie **cls, *cl;

  /* decrement the count of clauses in which key appears positively. */
  cls = el->pos;
  num_cls = el->pos_count;

  for (i = 0; i < num_cls; i++) {
    cl = cls[i];
    if (active_clause(cl)) {
      count = --count_of_clause(cl);
      if (count == 0) { 	
	/* restore what is done so far and then backtrack */
	count_of_clause(cl)++;
	for (j = 0; j < i; j++) 
	  if (active_clause(cls[j])) {
	    count_of_clause(cls[j])++;
	    cls[j]->psign++;
	  }
	if (save && GROW && Branch_open) trie2_record_conflict(cl);
	return handle_fail_end();
      } else if (count == 1) {
	/* store unit clauses in an array */
	Trie2_Ucl[Trie2_Ucl_idx++] = cl;
      }
      --cl->psign;
    }
  }

  /* remember the key for freezing its negative clauses */
  Tmp[--Tmp_idx_end] = key;
    
  /* remember the vars for backtracking */
  if (save) {
    el->next = Top_var2->next;
    Top_var2->next = el;
  }
  return SAT;
}

void trie2_clean_dependents (act, defreeze)
     struct var2 *act; int defreeze;
{
  struct var2 *it, *next;
  struct trie *cl, *cl2;
  
  /* at first, defreeze the clauses */
  if (defreeze) {
    it = act;
    while (it != NULL) {
      cl = frozen_cls(it);
      while (cl != CLAUSE3) {
	cl2 = frozen_lin(cl);
	frozen_lin(cl) = NULL;
	cl = cl2;
      }
      frozen_cls(it) = NULL;
      it = it->next;
    }
  }

  /* next, increase the counters */
  it = act->next;
  while (it != NULL) {
    trie2_clean_one_var(it);
    next = it->next;
    it->next = NULL;
    it = next;
  }
  trie2_clean_one_var(act);
  act->next = NULL;

  if (Old_cls != NULL) {
    cl = NHtrie;
    while (cl != Old_cls) {
      trie2_adjust_counts(cl);
#ifdef MORETRACE
      if (TRACE >= 9) {
	printf("Adjusted: ");
	print_clause_from_leaf(cl);
      }
#endif
      cl = cl->bro;
    }
    Old_cls = NULL;
  }
}

void trie2_clean_one_var (it)
     struct var2 *it; 
{
  int i, j, sign, num_cls;
  struct trie **cls, *cl;

  i = it->key;

#ifdef MORETRACE
  print_up_to_x;
#endif

  sign = Value[i];

  if (it->level == PURECHECK) {
    it->level = 0;
  } else if (sign) {  

    /* increment the counter of clauses */
    cls = it->neg;
    num_cls = it->neg_count;
    for (j = 0; j < num_cls; j++) 
      if (active_clause((cl=cls[j]))) {
#ifdef PRIVATE
	if (TRACE > 3 && trie2_cl_count(cl) < count_of_clause(cl)) {
	  print_clause_from_leaf(cl);
	  bug(2000);
	}
#endif
	count_of_clause(cl)++;
      }
  } else {
    cls = it->pos;
    num_cls = it->pos_count;
    for (j = 0; j < num_cls; j++) 
      if (active_clause((cl=cls[j]))) {
#ifdef PRIVATE
	if (TRACE > 3 && trie2_cl_count(cl) < count_of_clause(cl)) {
	  print_clause_from_leaf(cl);
	  bug(1000);
	}
#endif
	count_of_clause(cl)++;
	cl->psign++;
      }
  }

  Value[i] = DC;
}

trie2_handle_known_values ()
     /* search through Clause and collect all the unit clauses 
	into Trie2_Ucl. */
{
  if (Tmp_idx > MAX_ATOM) {
    fprintf ( stderr, "Tmp_idx (%d) > MAX_ATOM (%d).\n", Tmp_idx, MAX_ATOM);
    exit(1);
  }

  Tmp_idx_end = MAX_ATOM;
  Top_var2 = V2[0];
  while (Tmp_idx--) {
    if (trie2_propagate(V2[Tmp[Tmp_idx]], Value[Tmp[Tmp_idx]], 0) == UNSAT) 
      return UNSAT;
  }
  Tmp_idx = 0;

  return (trie2_handle_unit_clauses(0));
}

void trie2_freeze_clauses()
{
  int i, j, key, num_cls;
  struct var2 *el;
  struct trie **cls, *frozens, *cl;

  if (Tmp_idx_end < 0) {
    fprintf ( stderr, "Tmp is overflow (Tmp_idx_end = %d).\n", Tmp_idx_end);
    exit(1);
  }

  /* freeze clauses in which key appears as its value. */
  for (i = Tmp_idx_end; i < MAX_ATOM; i++) {
    el = V2[Tmp[i]];
    key = el->key;
    if (Value[key] == TT) {
      cls = el->pos;
      num_cls = el->pos_count;
    } else {
      cls = el->neg;
      num_cls = el->neg_count;
    }

    frozens = CLAUSE3;
    for (j = 0; j < num_cls; j++) if (active_clause((cl=cls[j]))) {
      frozen_lin(cl) = frozens;
      frozens = cl;
    }
    frozen_cls(el) = frozens;
  }
  Tmp_idx_end = MAX_ATOM;
}

int trie2_handle_unit_clauses (save)
     int save;
     /* handle each unit clause one by one ... */
{
  int i;
  struct trie *cl, *tr;

  /* At first, collect unit clauses from new clauses */
  if (Nonunit_cls != NHtrie) {
    cl = NHtrie; i = 0;
    while (cl != Nonunit_cls) {
      if (count_of_clause(cl) < 2) {
	i++;
	if (active_clause(cl)) {
	  /*print_clause_from_leaf(cl);*/
	  Trie2_Ucl[Trie2_Ucl_idx++] = cl;
	}
      }
      cl = cl->bro;
    }
    if (i == 0) Nonunit_cls = NHtrie;
  }

  if (Trie2_Ucl_idx >= MAX_LIST) {
    printf("Trie1_Ucl (%d) overflow!\n", MAX_LIST);
    exit(1);
  }

  while (Trie2_Ucl_idx > 0) {
    
    cl = Trie2_Ucl[--Trie2_Ucl_idx];
    /*printf("-");    print_clause_from_leaf(cl);*/
    
    if (active_clause(cl)) {
      if (count_of_clause(cl) == 0) {
	if (GROW && Backup_idx) trie2_record_conflict(cl);
	return handle_fail_end();
      } else {

	/* at first locate the literal which has no value */
	tr = cl->par; 
	i = 0;
	while (tr != NULL && i == 0) {
	  if (Value[tr->key] == DC) i = tr->key;
	  else tr = tr->par;
	}

	if (i) {
	  struct var2 *el = V2[i];

	  Value[i] = (cl->psign)? TT : FF;

#ifdef MORETRACE
	  print_x_has_to;
#endif

	  if (GROW) {
	    mark_of(el) = cl;
	    level_of(el) = Branch_open;
	  }
	  if (trie2_propagate(el, Value[i], save) == UNSAT) {
	
#ifdef MORETRACE
	    print_up_to_x;
#endif
	
	    Value[i] = DC;
	    return UNSAT;
	  }
	}
      }
    }
  }

  trie2_freeze_clauses();

  return SAT;
}

void trie2_record_conflict (cl)
     struct trie *cl;
{
  int i, j, u, uip, uip_num, total;
  int sign_arr[MAX_ATOM], uips[MAX_LENGTH], cl_arr[MAX_LENGTH];

  sign_arr[0] = 0;
  Conflict_cl = NULL;
  Old_cls = NHtrie;

  if (UIP) {
    Conflict_cl_level = (Backup[Backup_idx-1] > 0)? 0 : Branch_open;
    uip_num = trie2_locate_uips(cl, uips);
  } else {
    Conflict_cl_level = 0;
    uip = Backup[Backup_idx-1];
    if (uip < 0) uip = 0;
    uips[0] = uip;
    uip_num = 1;
  }

  for (u = 0; u < uip_num; u++) {
    total = trie2_collect_parents(cl, 0, cl_arr, uip=uips[u], u, sign_arr);

    for (i = sign_arr[0]; i > 0; i--) V2[sign_arr[i]]->pure = 0;

    if (total <= 0) {
      if (!UIP) Conflict_cl_level = Branch_open;
    } else {

      /* remove level 0 variables */
      /* fix the sign of literals in the new clause */
      for (i = 0; i < total; i++) {
	j = cl_arr[i];
	if (level_of(V2[j]) == 0) {
	  cl_arr[i--] = cl_arr[--total];
	} else {
	  V2[j]->pure = 0;
	  if (j != uip) {
	    if (!UIP && level_of(V2[j]) > Conflict_cl_level)  
	      Conflict_cl_level = level_of(V2[j]);
	  }
	  sign_arr[j] = NEG(Value[j]);
	}
      }

      if (u) {	
	cl_arr[total++] = j = uips[u-1];
	sign_arr[j] = Value[j];
	V2[j]->pure = 0;
      }

#ifdef MORETRACE
      if (TRACE > 4) {
	printf("\nCollected clause [open: %d]:\n   (", 
	       Branch_open);
	for (i = 0; i < total; i++) {
	  j = cl_arr[i];
	  printf(" %s%d[%d]", ((sign_arr[j])? "" : "-"), j, level_of(V2[j]));
	}
	printf(")\n\n");
      }
#endif

      /*check_cl(total, cl_arr, sign_arr);*/

      if (total <= GROW) {
	insert_sorter(cl_arr, total);

	/* integrating the new clause into the database */
	for (i = 0; i < total; i++) {
	  j = cl_arr[i];
	  if (sign_arr[j]) {
	    if (V2[j]->pos_extra <= 0) { 

#ifdef MORETRACE
	      if (TRACE > 2) 
		printf("  x%d has %d occurrences.\n", j, V2[j]->pos_count);
#endif

	      Miss_num++; 
	      if (trie2_add_extra_space(V2[j], TT) == 0) return; 
	    }
	  } else {
	    if (V2[j]->neg_extra <= 0) { 
	      Miss_num++; 

#ifdef MORETRACE
	      if (TRACE > 2) 
		printf(" -x%d has %d occurrences.\n", j, V2[j]->neg_count);
#endif

	      if (trie2_add_extra_space(V2[j], FF) == 0) {
		printf("WARNING: No space is available\n");
		return; 
	      }
	    }
	  }
	}

	if (total > 0) {

#ifdef MORETRACE
	  if (TRACE == 2) {
	    printf("\nInserting lemma [open: %d]:\n   (", 
		   Branch_open);
	    for (i = 0; i < total; i++) {
	      j = cl_arr[i];
	      printf(" %s%d[%d]", ((sign_arr[j])? "" : "-"), 
		     j, level_of(V2[j]));
	    }
	    printf(")\n\n");
	  }
#endif

	  cl = trie_insert_clause2(Root, cl_arr, sign_arr, total-1);
	  if (cl != CLAUSE3 && cl != NULL) {
	    trie2_integrate_clause(cl, (u)? uips[u-1] : 0);
	    if (uip_num > 1) Conflict_cl = NULL;
	  }
	} 
      }
    }
    cl = mark_of(V2[uip]);
  }
}

trie2_add_extra_space (it, sign)
     struct var2 *it;
     int sign;
{
  int i, old_space;
  struct trie **old_arr, **new_arr;

  if (sign) {
    old_space = it->pos_count;
    old_arr = it->pos;
  } else {
    old_space = it->neg_count;
    old_arr = it->neg;
  }

  /* new_arr = get_trie2_array(old_space+EXTRA_SPACE);*/
  new_arr = (struct trie **) malloc 
    ( sizeof ( struct trie *) * (old_space+EXTRA_SPACE));
  if (new_arr == NULL) {
    printf("Warning: ran out space; no more new clauses\n");
    return (GROW = 0);
  }
  Memory_count += (EXTRA_SPACE) * sizeof ( struct trie *);

  for (i = 0; i < old_space; i++) new_arr[i] = old_arr[i];
  free(old_arr);

  if (sign) {
    it->pos_extra = EXTRA_SPACE;
    it->pos = new_arr;
  } else {
    it->neg_extra = EXTRA_SPACE;
    it->neg = new_arr;
  }
  return 1;
}

struct trie *trie2_insert_clause (root, arr, signs, count)
     /* return the end of the clause */
     struct trie *root;
     int arr[], signs[], count;
{
  int total=count;
  int key;
  int sign;
  struct trie *no, *bro1, *bro2, *parent;

  parent = NULL;
  bro1 = bro2 = NULL;
  no = root;
  key = arr[count];
  sign = signs[key]; 

  while (no != NULL && count >= 0) {
    
    if (key == no->key) {
      if HAS_POS_CL(no) {
	
	if (sign == 1) {

#ifdef MORETRACE
	  print_x_subsume;
#endif

	  return NULL; /* skip this clause */
	} /* otherwise sign = 0 */

#ifdef MORETRACE
	print_x_skip;
#endif

	bro1 = no;
	no = no->bro; 
      } else if HAS_NEG_CL(no) {
      
	if (sign == 0) {

#ifdef MORETRACE
	  print_x_subsume;
#endif

	  return NULL; /* skip this clause */
	} /* otherwise sign == 1 */

#ifdef MORETRACE
	print_x_skip;
#endif

	bro1 = no;
	no = no->bro; 
      } else {
	parent = no;
	bro1 = NULL;
	if (sign) no = no->pos; else no = no->neg;
      }

      if (count-->0) {
	key = arr[count];
	sign = signs[key]; 
      }

    } else if (key < no->key) {
      bro2 = no;
      no = NULL;
    } else {
      bro1 = no;
      no = bro1->bro;
      while ((no != NULL) && (no->key < key)) {
	bro1 = no;
	no = bro1->bro;
      }
    }
  }

  if (count >= 0) {
    no = get_trie(); 
    no->key = arr[count]; 
    no->par = parent;
    if (parent != NULL)
      no->psign = signs[no->par->key];
    no->bro = bro2;
    if (bro1 == NULL) {
      if (parent != NULL) {
	if (signs[no->par->key])
	  no->par->pos = no;
	else 
	  no->par->neg = no;
      }
    } else {
      bro1->bro = no;
    }
    sign = signs[no->key];
    while (--count >= 0) {
      bro1 = no;
      no = get_trie();
      if (sign) bro1->pos = no;
      else bro1->neg = no;
      no->par = bro1;
      no->psign = sign;
      no->key = arr[count]; 
      sign = signs[no->key];
    }
  } else if (parent == NULL) {
    /* an empty clause is found */

#ifdef MORETRACE
    if (TRACE == 9) {
      printf("While inserting the clause:");
      print_clause(arr, signs, total);
    }
#endif

    return CLAUSE3;
  } else {
    no = parent;
  }

  if (signs[no->key]) {
    no->pos = CLAUSE3;
    no->neg = NULL;
  } else {
    no->neg = CLAUSE3;
    no->pos = NULL;
  }
  return no;
}

void trie2_integrate_clause (no, uip)
     struct trie *no;
{
  int psign, i;
  struct trie *cl;
  struct var2 *el;

  cl = get_trie();
  cl->par = no;
  if (uip) {
    count_of_clause(cl) = 1;
    cl->psign = (Value[uip])? 1 : 0;
    /* printf("You need to mark this clause!"); */
  } else 
    count_of_clause(cl) = cl->psign = 0;

  Conflict_cl = cl;
  cl->bro = NHtrie;
  NHtrie = cl;
  Clause_extra_num++;

  i = no->key;
  el = V2[i];
  if HAS_POS_CL(no) {
    el->pos_extra--;
    el->pos[el->pos_count++] = cl;
  } else {
    el->neg_extra--;
    el->neg[el->neg_count++] = cl;
  }
  
  psign = no->psign;
  no = no->par;
  while (no != NULL) {
    i = no->key;
    el = V2[i];
    if (psign) {
      el->pos_extra--;
      el->pos[el->pos_count++] = cl;
    } else {
      el->neg_extra--;
      el->neg[el->neg_count++] = cl;
    }
    psign = no->psign;
    no = no->par;
  }
}

trie2_collect_parents (cl, total, result, uip, cl_num, nodes)
     struct trie *cl;
     int total, result[], uip, nodes[];
{
  int i, idx;
  struct trie *stack[MAX_LENGTH];

  stack[0] = cl->par;
  idx = 1;

  while (idx) {
    cl = stack[--idx];
    if (cl->par != NULL) stack[idx++] = cl->par;
    i = cl->key;
    if (V2[i]->pure <= cl_num) {

      /* abort this procedure when one of the following is true */
      if ((Value[i] == DC) || (idx >= MAX_LENGTH) || (total >= MAX_LENGTH)) { 
	for (i = 0; i < total; i++)  V2[result[i]]->pure = 0;
	return 0; 
      }

      V2[i]->pure = cl_num+1;
      if (i != uip && level_of(V2[i]) >= Branch_open) {
	if (mark_of(V2[i]) != NULL) {
	  stack[idx++] = mark_of(V2[i])->par;
	  nodes[++nodes[0]] = i;
	} else {
	  result[total++] = i;
	}
      } else {
	result[total++] = i;
      }
    }
  }

  return total;
}

int trie2_locate_uips ( cl0, uips )
     struct trie *cl0;
     int uips[];
{
  int np[MAX_LENGTH], keys[MAX_LENGTH];
  int num, xyz;

  mark_of(V2[0]) = cl0;

  xyz = trie2_dfs( 0, -1, np, keys );
  if (xyz == 0) return 0;

  num = topo_sort(np, uips, xyz );
  while (xyz>=0) V2[keys[xyz--]]->pure = 0;

  if (num == 0) { uips[0] = 0; return 1; }
  return num;
}

int trie2_dfs ( key, dfsnum, np, keys )
     int key, dfsnum, np[], keys[];
{
  int  i;

  dfsnum++;
  if (dfsnum >= MAX_LENGTH) { 
    for (i = 0; i < dfsnum; i++) V2[keys[i]]->pure = 0; 
    return 0; 
  }
  V2[key]->pure = -dfsnum;
  np[dfsnum] = 0;
  keys[dfsnum] = key;

  if (mark_of(V2[key]) != NULL) {
    struct trie *cl;

    cl = mark_of(V2[key])->par;
    while (cl != NULL) {
      if ((i= cl->key) != key) {
	if (level_of(V2[i]) >= Branch_open) {
	  if (V2[i]->pure >= 0 && 
	      (dfsnum = trie2_dfs( i, dfsnum, np, keys )) == 0) {
	    return 0;
	  }
	  ++np[-V2[i]->pure];
	} else {
	  if (Conflict_cl_level < level_of(V2[i])) 
	    Conflict_cl_level = level_of(V2[i]);
	}
      }
      cl = cl->par;
    }
  }
  
  return dfsnum;
}

int topo_sort (np, uips, total )
     int np[], uips[], total;
{
  int uipnum, i, key, idx, idx2, last_dfsnum;
  int stack[MAX_LENGTH], inlevel[MAX_LENGTH], level[MAX_LENGTH];
  int lvl, dfsnum, d2;
  struct trie *cl;

  /* initialization */
  stack[0] = uipnum = last_dfsnum = 0;
  for (i = 0; i <= total; i++) inlevel[i] = level[i] = 0;
  idx = 1;
  idx2 = MAX_LENGTH;

  while (idx) {
    key = stack[--idx];
    dfsnum = -V2[key]->pure;
    lvl = level[dfsnum];

    while (idx2 < MAX_LENGTH && np[stack[idx2]] == 0) idx2++;

    if (idx == 0 &&  /* no other roots */
	idx2 >= MAX_LENGTH &&  /* no other paths */
	inlevel[lvl] == 0) {  /* no other paths */
      
      if (last_dfsnum < dfsnum) {
	uips[uipnum++] = key;
	last_dfsnum = dfsnum+1; 
	UIP_num++;

#ifdef MORETRACE
	if (TRACE > 3) printf("     %d is a UIP\n", key);
#endif
      }
    }

    inlevel[lvl] = 1;

    if (cl = mark_of(V2[key])) {

#ifdef MORETRACE
      if (TRACE > 3) printf("   Children of %d (level=%d): [", key, lvl);
#endif

      cl = cl->par;
      while (cl != NULL) {
	if ((i= cl->key) != key &&
	    level_of(V2[i]) >= Branch_open) {

	  d2 = -V2[i]->pure;

#ifdef MORETRACE
	  if (TRACE > 3) printf(" %d (level=%d)", i, level[d2]);
#endif	  

	  if (lvl >= level[d2]) {
	    level[d2] = lvl+1;
	  }
	  if (--np[d2] == 0) stack[idx++] = i;
	  else stack[--idx2] = d2;
	  if (idx >= idx2) { bug(MAX_LENGTH); return 0; }
	} 
	cl = cl->par;
      }

#ifdef MORETRACE
      if (TRACE > 3) printf(" ]\n");
#endif
    }
  }

  if (uipnum) {
    if (uips[uipnum-1] != key) 
      uips[uipnum++] = key;
    else
      UIP_num--;
  }

  return uipnum;
}

int trie2_max_level (cl, key)
     struct trie *cl;
     int key;
{
  int max = 0;
  cl = cl->par;
  while (cl != NULL) {
    if (cl->key != key &&
	level_of(V2[cl->key]) > max)
      max = level_of(V2[cl->key]);
    cl = cl->par;
 }
  return max;
}

void trie2_adjust_counts (cl)
     struct trie *cl;
{
  struct trie *tr;
  struct trie *son;
  int count, pcount;

  count = pcount = 0;

  tr = cl->par;

  if (Value[tr->key] == DC) {
    count++;
    if (HAS_POS_CL(tr)) pcount++;
  }

  son = tr; 
  tr = tr->par;
  while (tr != NULL) {
    if (Value[tr->key] == DC) {
      count++;
      if (son->psign) pcount++;
    }
    son = tr;
    tr = tr->par;
  }

  count_of_clause(cl) = count;
  cl->psign = pcount;
}

int trie2_cl_count (cl)
     struct trie *cl;
{
  struct trie *tr;
  struct trie *son;
  int count;

  count = 0;

  tr = cl->par;
  if (Value[tr->key] == DC) {
    count++;
  }
  if (HAS_POS_CL(tr) && Value[tr->key] != FF) return 100;
  if (HAS_NEG_CL(tr) && Value[tr->key] == FF) return 100;

  son = tr; 
  tr = tr->par;
  while (tr != NULL) {
    if (Value[tr->key] == DC) {
      count++;
    } else if (son->psign == 0 && Value[tr->key] == FF) return 100;
    else if (son->psign == 1 && Value[tr->key] != FF) return 100;
    son = tr;
    tr = tr->par;
  }

  return count;
}

void trie2_empty (cl)
     struct trie *cl;
{
  while (cl != NULL) {
    if (trie2_is_empty(cl)) print_clause_from_leaf(cl);
    cl = cl->bro;
  }
}

int trie2_is_empty (tr)
     struct trie *tr;
{
  struct trie *son;

  tr = tr->par;
  if (HAS_POS_CL(tr) && Value[tr->key] != FF) return 0;
  if (HAS_NEG_CL(tr) && Value[tr->key] == FF) return 0;

  son = tr;
  tr = tr->par;
  while (tr != NULL) {
    if (son->psign && Value[tr->key] != FF) return 0;
    if (son->psign == 0 && Value[tr->key] == FF) return 0;
    son = tr; 
    tr = tr->par;
  }
  return 1;
}

#ifdef MORETRACE
check_cl (total, arr, sign)
     int total, arr[], sign[];
{
  int pos[201];
  static int model[201];
  int i, j;

  if (total == 0) {
    j = 0;
    for (i = 1; i <= 200; i++) model[i] = FF;
    for (i = 1; i <= 200; i++) {
      if (pos[j] == i) { model[i] = TT; j++; }
    }
    return 0;
  } else {
    for (i = 0; i < total; i++) 
      if (sign[arr[i]] == model[arr[i]]) return 0;
    return 1;
  }
}
#endif

void trie2_stats ()
{
  printf("\n");
  if (Clause_extra_num) 
    printf("The number of created clauses is %d (%d array updates).\n",
	   Clause_extra_num, Miss_num);
  if (UIP_num)
    printf("The number of found UIPs is %d.\n",
	   UIP_num);
  if (Jump_num)
    printf("The number of backjumping is %d.\n",
	   Jump_num);
  if (Pure_num)
    printf("The number of pure deletions is %d.\n",
	   Pure_num);
}

void print_all_clauses (cl)
     struct trie *cl;
{
  while (cl != NULL) {
    print_clause_from_leaf(cl);
    cl = cl->bro;
  }
}

int trie2_insert_one_key (key, key_arr, total)
     int key, key_arr[], total;
     /* insert a literal into the current clause.
	return 0 if the key is in key_arr.
	otherwise, return the length of the current clause. */
{
  int j;

  for (j = 0; j < total; j++)
    if (key == key_arr[j]) j = MAX_ATOM;

  if (j >= MAX_ATOM) return 0;
  key_arr[total++] = key;
  return total;
}

void trie2_record_keeping ()
{
  int i;

  /* adjust splitting strategies */
  SELECT = 2;
  if (NH_num > 2) {
    i = (100*Clause_num)/(NH_num-2);
    /* if NH > 20% select 1. */
    if (i < 350) SELECT = 1;
    else if (i >= 4242) {
      /* if NH < 2% select 3. */
      if (i == 6233 || i == 5187) SELECT = 3;
      i = CHOICE1;
      CHOICE1 = CHOICE2;
      CHOICE2 = i;
    }
  }
}

int weight_of_clauses ()
     /* return the number of false clauses under the current 
	interpretation */
{
  return count_false_cl_trie(Root);
}

int count_false_cl_trie (tr)
     struct trie *tr;
{
  if (tr == CLAUSE3) {
    return 1;
  } else {
    int num_f_cl = 0;
    while (tr != NULL) {
      if (Value[tr->key] == FF) {
	if (tr->pos != NULL)
	  num_f_cl += count_false_cl_trie(tr->pos);
      } else if (tr->neg != NULL) {
	  num_f_cl += count_false_cl_trie(tr->neg);
      }
      tr = tr->bro;
    }
    return num_f_cl;
  } 
}

